from .predict import predict
from .shap_vals import shap_values
from .annotate import annotate

import pandas as pd


def isv(cnvs, proba: bool = True, shap: bool = False):
    """Annotate and Predict pathogenicity of CNVs

    :param cnvs: a list, np.array or pandas dataframe with 4 columns representing chromosome (eg, chr3),
    cnv start (grch38), cnv end (grch38) and cnv_type (DUP or DEL)
    :param proba: whether probabilities should be returned
    :param shap: whether shap values should be calculated
    :return: ISV output as a pandas dataframe
    """
    final = cnvs.copy()
    annotated = annotate(cnvs)

    # 1. Add predictions
    final["ISV"] = predict(annotated, proba)

    # 2. Add SHAP values
    if shap:
        svs = shap_values(annotated)

        final = pd.concat([final, svs], axis=1)

    return final



import shap
import numpy as np
import pandas as pd

from isv.config import settings
from isv.scripts.prepare_df import prepare
from isv.scripts.open_model import open_model
from isv.scripts.constants import HUMAN_READABLE, LOSS_ATTRIBUTES, GAIN_ATTRIBUTES


def shap_values_with_same_cnv_type(annotated_cnvs: pd.DataFrame, cnv_type: str):
    """Calculate SHAP values for CNVs with the same cnv type

    :param annotated_cnvs: Raw counts of genomic elements
    :param cnv_type: type of cnv
    :return: explainer object
    """
    X, X_train = prepare(annotated_cnvs, cnv_type, return_train=True)
    model = open_model(settings.model_dir + f'ISV_{cnv_type}.json')

    explainer = shap.TreeExplainer(
        model,
        X_train,
        model_output='probability')

    return explainer(X).values


def shap_values(annotated_cnvs: pd.DataFrame):
    """Calculate SHAP values

    :param annotated_cnvs:
    :return: explainer object
    """
    del_ind = np.where(annotated_cnvs.cnv_type == "DEL")[0]
    dup_ind = np.where(annotated_cnvs.cnv_type == "DUP")[0]

    res = []

    if del_ind.shape[0] > 0:
        attributes = LOSS_ATTRIBUTES
        hr_attributes = ['SHAP_' + HUMAN_READABLE[i].replace(' ', '_') for i in attributes]
        sv = shap_values_with_same_cnv_type(annotated_cnvs.iloc[del_ind], "loss")
        res.append(pd.DataFrame(sv, columns=hr_attributes))

    if dup_ind.shape[0] > 0:
        attributes = GAIN_ATTRIBUTES
        hr_attributes = ['SHAP_' + HUMAN_READABLE[i].replace(' ', '_') for i in attributes]
        sv = shap_values_with_same_cnv_type(annotated_cnvs.iloc[dup_ind], "gain")
        res.append(pd.DataFrame(sv, columns=hr_attributes))

    res = pd.concat(res)
    res.index = np.concatenate([del_ind, dup_ind])
    res = res.sort_index()

    return res



import pathlib


class Settings:
    def __init__(self):
        root_dir = str(pathlib.Path(__file__).parent.absolute())
        self.model_dir = root_dir + '/models/'
        self.data_dir = root_dir + '/data/'
        self.valid_chromosomes = [f'chr{i}' for i in range(1, 23)] + ['chrX', 'chrY']
        self.chromosome_dict = dict(zip(self.valid_chromosomes, range(1, 25)))

        self.gene_type_dict = {'protein_coding': 0, 'pseudogene': 1, 'lncRNA': 2,
                               'miRNA': 3, 'rRNA': 4, 'snRNA': 5, 'unevaluated': 6}

        self.regulatory_type_dict = {'CTCF_binding_site': 7, 'TF_binding_site': 8,
                                     'curated': 9, 'enhancer': 10, 'open_chromatin_region': 11,
                                     'promoter': 12, 'promoter_flanking_region': 13}
        
        self.attributes = [
            'gencode_genes',
            'protein_coding',
            'pseudogenes',
            'mirna',
            'lncrna',
            'rrna',
            'snrna',
            'morbid_genes',
            'disease_associated_genes',
            'hi_genes',  # ONLY FOR LOSSES
            'regions_HI',  # ONLY FOR LOSSES
            'regions_TS',  # ONLY FOR GAINS
            'regulatory',
            'regulatory_enhancer',
            'regulatory_open_chromatin_region',
            'regulatory_promoter',
            'regulatory_promoter_flanking_region',
            'regulatory_ctcf_binding_site',
            'regulatory_tf_binding_site',
            'regulatory_curated'
        ]


settings = Settings()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
import numba as nb
import pandas as pd
import os
import gzip
import json
import time

from isv.config import settings


def open_data(f):
    """Open preprocessed data
    :param f: filepath
    :return: python dictionary of numpy arrays
    """
    with gzip.open(f, 'r') as g:
        temp = json.loads(g.read().decode("utf-8"))
    
    cd = {i:None for i in range(1, 25)} # just a placeholder 
    # as np arrays -> for numba
    for i in range(1, 25):
        cd[i] = np.array(temp[str(i)], dtype=np.int64)
    
    return cd


@nb.jit(nopython=True)
def get_el(arr, start: int, end: int):
    """Get overlapped elements
    :param arr: array to query
    :param start: start position
    :param end: end position
    :return: rows of array overlapped by start, end positions
    """
    return arr[np.where((arr[:, 1] <= end) & (arr[:, 2] >= start))[0]]


def annotate_cnv(chrom, start, end, gencode_genes, regulatory, hi_genes, hits_regions):
    """Annotate a candidate CNV

    :param chrom: chromosome identifier, eg. "chr1"
    :param start: start position on the GRCh38 assembly
    :param end: end position on the GRCh38 assembly
    :param gencode_genes: preprocessed gencode genes dict
    :param regulatory: preprocessed regulatory genes dict
    :param hi_genes: preprocessed hi_genes genes dict
    :param hi_regions: preprocessed hi_regions genes dict
    :param ts_regions: preprocessed ts_regions genes dict
    :return: annotation
    """
    
    overlapped_genes = get_el(gencode_genes[chrom], start, end)
    overlapped_regulatory = get_el(regulatory[chrom], start, end)
    overlapped_hi_genes = get_el(hi_genes[chrom], start, end)
    overlapped_hits_regions = get_el(hits_regions[chrom], start, end)
    
    return np.array([
        len(overlapped_genes),  # gencode_genes
        np.sum(overlapped_genes[:, 3] == settings.gene_type_dict["protein_coding"]),
        np.sum(overlapped_genes[:, 3] == settings.gene_type_dict["pseudogene"]),
        np.sum(overlapped_genes[:, 3] == settings.gene_type_dict["miRNA"]),
        np.sum(overlapped_genes[:, 3] == settings.gene_type_dict["lncRNA"]),
        np.sum(overlapped_genes[:, 3] == settings.gene_type_dict["rRNA"]),
        np.sum(overlapped_genes[:, 3] == settings.gene_type_dict["snRNA"]),
        np.sum(overlapped_genes[:, 4]),  # morbid genes
        np.sum(overlapped_genes[:, 5]),  # disease associated genes
        len(overlapped_hi_genes),
        np.sum(overlapped_hits_regions[:, 3]),  # HI regions
        np.sum(overlapped_hits_regions[:, 4]),  # TS regions
        len(overlapped_regulatory),  # all overlapped regulatory elements
        np.sum(overlapped_regulatory[:, 3] == settings.regulatory_type_dict["enhancer"]),
        np.sum(overlapped_regulatory[:, 3] == settings.regulatory_type_dict["open_chromatin_region"]),
        np.sum(overlapped_regulatory[:, 3] == settings.regulatory_type_dict["promoter"]),
        np.sum(overlapped_regulatory[:, 3] == settings.regulatory_type_dict["promoter_flanking_region"]),
        np.sum(overlapped_regulatory[:, 3] == settings.regulatory_type_dict["CTCF_binding_site"]),
        np.sum(overlapped_regulatory[:, 3] == settings.regulatory_type_dict["TF_binding_site"]),
        np.sum(overlapped_regulatory[:, 3] == settings.regulatory_type_dict["curated"])
    ], dtype=np.int32)


# %%
def annotate(cnvs):
    """
    :param cnvs: a list, np.array or pandas dataframe with 4 columns representing chromosome (eg, chr3),
    cnv start (grch38), cnv end (grch38) and cnv_type (DUP or DEL)
    :return: pd DataFrame of annotated CNVs
    """
    if isinstance(cnvs, list) or isinstance(cnvs, np.ndarray):
        cnvs = pd.DataFrame(cnvs)
    
    assert isinstance(cnvs, pd.core.frame.DataFrame),\
        "Please supply a list, np.array or pandas dataframe with 4 columns\
            representing the chromosome id, start (grch38), end (grch38)\
                and cnv_type (eiter DUP or DEL)"
    
    assert cnvs.shape[1] == 4, "Please supply dataframe with 4 columns only"

    cnvs.columns = ["chrom", "start", "end", "cnv_type"]

    assert np.sum([i in ["DEL", "DUP"] for i in cnvs.cnv_type]) == len(cnvs), "cnv type has to be either DUP or DEL"

    # Just in case something is wrong with indexes
    cnvs.reset_index(inplace=True, drop=True)

    # Make sure that chromosomes are in the right format
    cnvs.chrom = [i if i.startswith("chr") else f"chr{i}" for i in cnvs.chrom]

    # Load databases
    print("Loading databases")
    gencode_genes = open_data(os.path.join(settings.data_dir, "preprocessed", "gencode_genes.json.gz"))
    regulatory = open_data(os.path.join(settings.data_dir, "preprocessed", "regulatory.json.gz"))
    hi_genes = open_data(os.path.join(settings.data_dir, "preprocessed", "hi_genes.json.gz"))
    hits_regions = open_data(os.path.join(settings.data_dir, "preprocessed", "hits_regions.json.gz"))
    
    # annotate cnvs
    print("Annotating CNVs")
    annotated = np.empty((cnvs.shape[0], len(settings.attributes)), dtype=np.int64)
    start = time.time()
    for i, cnv in cnvs.iterrows():
        annotated[i] = annotate_cnv(settings.chromosome_dict[cnv.chrom],
                                    cnv.start, cnv.end, gencode_genes,
                                    regulatory, hi_genes, hits_regions)
    
    elapsed = round(time.time() - start, 9)
    print(f"Annotated in {elapsed} seconds")
    
    return pd.concat([cnvs, pd.DataFrame(annotated, columns=settings.attributes)], axis=1)



import xgboost as xgb

from isv.config import settings
from isv.scripts.prepare_df import prepare
from isv.scripts.open_model import open_model

import numpy as np
import pandas as pd


def predict_with_same_cnv_type(annotated_cnvs: pd.DataFrame, cnv_type: str, proba: bool = True):
    """Return model predictions for a selected dataframe
`
    :param annotated_cnvs: Raw counts of genomic elements
    :param cnv_type: type of cnv
    :param proba: return probabilities
    :return: yhat: predicted values
    """
    model = open_model(settings.model_dir + f'ISV_{cnv_type}.json')
    X = prepare(annotated_cnvs, cnv_type)

    if proba:
        if isinstance(model, xgb.core.Booster):
            X_dmat = xgb.DMatrix(X)
            yhat = model.predict(X_dmat)
        else:
            yhat = model.predict_proba(X)[:, 1]

    else:
        if isinstance(model, xgb.core.Booster):
            X_dmat = xgb.DMatrix(X)
            yhat = model.predict(X_dmat)
            yhat = (yhat > 0.5) * 1
        else:
            yhat = model.predict(X)

    return yhat


def predict(annotated_cnvs: pd.DataFrame, proba: bool = True):
    """Predict bulk of CNVs with different cnv types

    :param annotated_cnvs: Annotated CNVs
    :param proba: whether probabilities should be calculated
    :return: predictions
    """

    del_ind = np.where(annotated_cnvs.cnv_type == "DEL")[0]
    dup_ind = np.where(annotated_cnvs.cnv_type == "DUP")[0]

    yh = np.empty(len(annotated_cnvs), dtype=[np.int8, np.float64][proba * 1])
    if len(del_ind) > 0:
        yh[del_ind] = predict_with_same_cnv_type(annotated_cnvs.iloc[del_ind], "loss", proba)

    if len(dup_ind) > 0:
        yh[dup_ind] = predict_with_same_cnv_type(annotated_cnvs.iloc[dup_ind], "gain", proba)

    return yh
